Imports System.Collections
Imports System.Windows.Forms
Imports System.Drawing

Namespace TileBasisFormat

    Public Enum Direction
        Left_Direction = 0
        Up_Direction
        Right_Direction
        Down_Direction
        No_Direction
    End Enum
    Public Structure TileLocation

        Friend TX As Integer
        Friend TY As Integer

        Public Sub New(ByVal X As Integer, ByVal Y As Integer)
            'Returns a tilelocation with the specified location.
            Me.TX = X
            Me.TY = Y
        End Sub

        Public Shared Function LeftFace(ByVal D As Direction) As Direction
            'Returns a specified direction's left face stance (90 counterclockwise)
            If D = Direction.Left_Direction Then
                Return Direction.Down_Direction
            ElseIf D = Direction.Up_Direction Then
                Return Direction.Left_Direction
            ElseIf D = Direction.Right_Direction Then
                Return Direction.Up_Direction
            ElseIf D = Direction.Down_Direction Then
                Return Direction.Right_Direction
            End If
        End Function

        Public Shared Function RightFace(ByVal D As Direction) As Direction
            'Returns a specified direction's right face stance (90 clockwise)
            If D = Direction.Left_Direction Then
                Return Direction.Up_Direction
            ElseIf D = Direction.Up_Direction Then
                Return Direction.Right_Direction
            ElseIf D = Direction.Right_Direction Then
                Return Direction.Down_Direction
            ElseIf D = Direction.Down_Direction Then
                Return Direction.Left_Direction
            End If
        End Function

        Public Shared Function AboutFace(ByVal D As Direction) As Direction
            'Returns a specified direction's about face stance (180)
            If D = Direction.Left_Direction Then
                Return Direction.Right_Direction
            ElseIf D = Direction.Up_Direction Then
                Return Direction.Down_Direction
            ElseIf D = Direction.Right_Direction Then
                Return Direction.Left_Direction
            ElseIf D = Direction.Down_Direction Then
                Return Direction.Up_Direction
            End If
        End Function

        Public Sub SetByValues(ByVal X As Integer, ByVal Y As Integer)
            'Sets the location of this instance.
            TX = X
            TY = Y
        End Sub

        Public Function Translation(ByVal delX As Integer, ByVal delY As Integer) As TileLocation
            'Returns a tile with the specified displacement.
            Translation.TX = Me.TX + delX
            Translation.TY = Me.TY + delY
        End Function

        Public Overloads Function MoveByDirection(ByVal D As Direction) As TileLocation
            'Moves this instance one step in the specified direction.
            Dim Direc As TileLocation = Me
            If D = Direction.Left_Direction Then
                Direc.TX = Me.TX - 1

            ElseIf D = Direction.Up_Direction Then
                Direc.TY = Me.TY - 1

            ElseIf D = Direction.Right_Direction Then
                Direc.TX = Me.TX + 1

            ElseIf D = Direction.Down_Direction Then
                Direc.TY = Me.TY + 1

            End If

            Return Direc
        End Function

        Public Overloads Function MoveByDirection(ByVal D As Direction, ByVal Distance As Integer) As TileLocation
            'Moves this instance along a direction by a distance.
            MoveByDirection = Me
            If D = Direction.Left_Direction Then
                MoveByDirection.TX = Me.TX - Distance

            ElseIf D = Direction.Up_Direction Then
                MoveByDirection.TY = Me.TY - Distance

            ElseIf D = Direction.Right_Direction Then
                MoveByDirection.TX = Me.TX + Distance

            ElseIf D = Direction.Down_Direction Then
                MoveByDirection.TY = Me.TY + Distance

            End If
        End Function

        Public Function InBounds(ByRef Boundary As System.Drawing.Rectangle) As Boolean
            'Returns true if this instance is inside the bounds of the specified rectangle object.
            Return (Me.TX >= Boundary.Left AndAlso Me.TX < Boundary.Right AndAlso Me.TY >= Boundary.Top AndAlso Me.TY < Boundary.Height)
        End Function

        Public Function MoveIntoBounds(ByRef Boundary As System.Drawing.Rectangle) As Integer
            Dim Disp As Integer = 0
            'Moves this instance into the bounds of a specified rectangle object
            If Me.TX < Boundary.Left Then
                Disp = Boundary.Left - Me.TX
                Me.TX = Boundary.Left
            ElseIf Me.TX >= Boundary.Right Then
                Disp = Me.TX - Boundary.Right + 1
                Me.TX = Boundary.Right - 1
            End If
            If Me.TY < Boundary.Top Then
                Disp += Boundary.Top - Me.TY
                Me.TY = Boundary.Top
            ElseIf Me.TY >= Boundary.Bottom Then
                Disp += Me.TY - Boundary.Bottom + 1
                Me.TY = Boundary.Bottom - 1
            End If
            Return Disp
        End Function

        Public Function DirectionalDistance(ByVal D As Direction, ByVal MoveTo As TileLocation) As Integer
            Select Case D
                Case Direction.Down_Direction
                    Return MoveTo.TY - Me.TY
                Case Direction.Left_Direction
                    Return Me.TX - MoveTo.TX
                Case Direction.Right_Direction
                    Return MoveTo.TX - Me.TX
                Case Direction.Up_Direction
                    Return Me.TY - MoveTo.TY
                Case Else
                    Return Integer.MaxValue
            End Select
        End Function

        Public Shared Function KeyCodeFromDirection(ByVal D As Direction) As System.Windows.Forms.Keys
            Select Case D
                Case Direction.Down_Direction
                    Return Keys.Down
                Case Direction.Left_Direction
                    Return Keys.Left
                Case Direction.Right_Direction
                    Return Keys.Right
                Case Direction.Up_Direction
                    Return Keys.Up
                Case Else
                    Return Keys.Enter
            End Select
        End Function
        Public Shared Function DirectionFromKeyCode(ByVal K As System.Windows.Forms.Keys) As Direction
            Select Case K
                Case Keys.Down
                    Return Direction.Down_Direction
                Case Keys.Left
                    Return Direction.Left_Direction
                Case Keys.Right
                    Return Direction.Right_Direction
                Case Keys.Up
                    Return Direction.Up_Direction
                Case Else
                    Return Direction.No_Direction
            End Select
        End Function

        Public Overloads Function Equals(ByVal TL As TileLocation) As Boolean
            Return TL.TX = Me.TX AndAlso TL.TY = Me.TY
        End Function

    End Structure
    Interface ITile
        ReadOnly Property XLocation() As Integer
        ReadOnly Property YLocation() As Integer
        ReadOnly Property Tilelocation() As Tilelocation
    End Interface
    Public Class BackgroundTile

        Implements ITile, IComparable


        Protected mTL As Tilelocation
        Private mData As Short

        Public Property Data() As Short
            Get
                Return mData
            End Get
            Set(ByVal Value As Short)
                mData = Value
            End Set
        End Property
        Public Sub SwapData(ByRef SecondTile As BackgroundTile)
            Dim D As Short = mData
            mData = SecondTile.mData
            SecondTile.mData = D
        End Sub
        Public Sub New(ByVal X As Integer, ByVal Y As Integer)
            mTL.TX = X
            mTL.TY = Y
        End Sub
        Public Sub New(ByVal TileLoc As Tilelocation)
            mTL = TileLoc
        End Sub
        'Public Function XLocation() As Integer Implements ITile.XLocation
        'End Function

        'Public Function YLocation() As Integer Implements ITile.YLocation
        'End Function

        'Public Function Tilelocation() As TileLocation Implements ITile.TileLocation
        'End Function
#Region "Implements ITile"
        Public ReadOnly Property Tilelocation() As Tilelocation Implements ITile.Tilelocation
            Get
                Return mTL

            End Get
        End Property

        Public ReadOnly Property XLocation() As Integer Implements ITile.XLocation
            Get
                Return mTL.TX

            End Get
        End Property

        Public ReadOnly Property YLocation() As Integer Implements ITile.YLocation
            Get
                Return mTL.TY

            End Get
        End Property
#End Region

        Public Function CompareTo(ByVal obj As Object) As Integer Implements System.IComparable.CompareTo
            Dim T As BackgroundTile = DirectCast(obj, BackgroundTile)
            Return Me.mTL.TY.CompareTo(T.mTL.TY)
        End Function

        Public Function At(ByVal X As Integer, ByVal Y As Integer) As Boolean
            Return (Me.mTL.TX = X AndAlso Me.mTL.TY = Y)
        End Function


    End Class
    Public Class MapCursor
        Inherits BackgroundTile
        Private S1 As Short
        Private S2 As Short
        Private Head As Direction

        Protected MyBounds As System.Drawing.Rectangle


        Public Sub New(ByVal StartX As Integer, ByVal StartY As Integer, ByRef Boundary As System.Drawing.Rectangle)
            MyBase.New(StartX, StartY)
            MyBounds = Boundary
            Head = Direction.Left_Direction
        End Sub
        Public Sub New(ByVal StartX As Integer, ByVal StartY As Integer, ByVal Data As Short, ByRef Boundary As System.Drawing.Rectangle)
            MyBase.New(StartX, StartY)
            MyBounds = Boundary
            Me.Data = Data
            Head = Direction.Left_Direction
        End Sub
        Public Sub New(ByVal StartX As Integer, ByVal StartY As Integer, ByVal Data As Short, ByRef Boundary As System.Drawing.Rectangle, ByVal S1 As Short, ByVal S2 As Short)
            MyBase.New(StartX, StartY)
            MyBounds = Boundary
            Me.Data = Data
            Head = Direction.Left_Direction
            Me.S1 = S1
            Me.S2 = S2
        End Sub
        Public Sub New(ByVal BR As IO.BinaryReader, ByVal Boundary As System.Drawing.Rectangle)
            MyBase.New(0, 0)
            Dim TL As TileLocation = New TileLocation(BR.ReadInt32(), BR.ReadInt32())
            MyBounds = Boundary
            MyBase.mTL = TL
            MyBase.Data = BR.ReadInt16()
            S1 = BR.ReadInt16()
            S2 = BR.ReadInt16()
        End Sub

        Public Overloads Function Move(ByVal D As Direction) As Integer
            'Moves the cursor forward by one within its boundary.
            Dim TL As TileLocation = Me.Tilelocation()
            Dim Dst As Integer
            TL = TL.MoveByDirection(D)
            'Move into direction
            TL.MoveIntoBounds(MyBounds)
            'If out of bounds then move back into bounds
            Dst = Me.mTL.DirectionalDistance(D, TL)
            'Record the distance traveled
            Me.mTL = TL
            'Set the location
            Head = D
            Return Dst
        End Function
        Public Overloads Function Move(ByVal D As Direction, ByVal Distance As Integer) As Integer
            'Moves the cursor forward by the specified amount within its boundary.
            Dim TL As TileLocation = Me.Tilelocation()
            Dim Dst As Integer
            TL = TL.MoveByDirection(D, Distance)
            'Move into direction by the specified amount
            TL.MoveIntoBounds(MyBounds)
            'If out of bounds then move back into bounds
            Dst = Me.mTL.DirectionalDistance(D, TL)
            'Record the distance traveled.
            Me.mTL = TL
            'Set the location.
            Head = D
            Return Dst
        End Function
        Public Overloads Function Move(ByVal D As Direction, ByRef Map As MapGrid, ByVal BlockedData() As Short) As Integer
            'Moves the cursor forward by one if path on the map is not blocked.
            Dim TL As TileLocation = Me.Tilelocation()
            Dim St As Short
            Dim Dst As Integer
            TL = TL.MoveByDirection(D)
            'Move into direction by the specified amount.
            If Not TL.InBounds(MyBounds) Then
                TL.MoveIntoBounds(MyBounds)
                'Move back if we are not in bounds
            Else
                St = Map.TileData(TL.TX, TL.TY)
                'Get current tile info.
                If Array.BinarySearch(BlockedData, St) >= 0 Then
                    'If tile is blocked
                    TL = TL.MoveByDirection(Tilelocation.AboutFace(D))
                    'Move back.
                End If
            End If
            Dst = Me.mTL.DirectionalDistance(D, TL)
            Me.mTL = TL
            Head = D
            Return Dst
        End Function
        Public Overloads Function Move(ByVal D As Direction, ByVal Distance As Integer, ByRef Map As MapGrid, ByVal BlockedData() As Short, ByVal FullPathCheck As Boolean) As Integer
            Dim TL As TileLocation = Me.Tilelocation()
            Dim Dst As Integer
            For U As Integer = 1 To Distance
                TL = TL.MoveByDirection(D)
                'Move forward for each step.
                If Not TL.InBounds(MyBounds) OrElse Array.BinarySearch(BlockedData, Map.TileData(Me.mTL.TX, Me.mTL.TY)) < 0 Then
                    'If out of bounds or on a blocked tile at any step...
                    TL = TL.MoveByDirection(Tilelocation.AboutFace(D))
                    'Move back to the previous step.
                    Dst = U - 1
                    'Set the distance
                    U = Distance
                    'And kill the loop variable.
                End If
            Next
            Me.mTL = TL
            Head = D
            Return Dst
        End Function

        Public Sub MoveTo(ByVal TL As Tilelocation)
            MyBase.mTL = TL
        End Sub

        Public Function MoveAhead() As Integer
            'Moves the cursor forward by one within its boundary.
            Dim TL As TileLocation = Me.Tilelocation()
            Dim Dst As Integer
            TL = TL.MoveByDirection(Head)
            'Move into direction
            TL.MoveIntoBounds(MyBounds)
            'If out of bounds then move back into bounds
            Dst = Me.mTL.DirectionalDistance(Head, TL)
            'Record the distance traveled
            Me.mTL = TL
            'Set the location
            Return Dst

        End Function
        Public Function MoveBack() As Integer
            'Moves the cursor forward by one within its boundary.
            Dim TL As TileLocation = Me.Tilelocation()
            Dim Dst As Integer
            Head = Tilelocation.AboutFace(Head)
            TL = TL.MoveByDirection(Head)
            'Move into direction
            TL.MoveIntoBounds(MyBounds)
            'If out of bounds then move back into bounds
            Dst = Me.mTL.DirectionalDistance(Head, TL)
            'Record the distance traveled
            Me.mTL = TL
            'Set the location
            Return Dst

        End Function

        Public Function IsInTileCollection(ByVal TLS As BackgroundTile()) As Boolean
            Dim LV As Integer
            For LV = 0 To TLS.GetUpperBound(0)
                If TLS(LV).Tilelocation.Equals(MyBase.Tilelocation()) Then
                    IsInTileCollection = True
                End If
            Next
        End Function

        Public ReadOnly Property Boundary() As Rectangle
            Get
                Return MyBounds
            End Get
        End Property

        Public Property Setting1() As Short
            Get
                Return S1
            End Get
            Set(ByVal Value As Short)
                S1 = Value
            End Set
        End Property
        Public Property Setting2() As Short
            Get
                Return S2
            End Get
            Set(ByVal Value As Short)
                S2 = Value
            End Set
        End Property
        Public Property Heading() As Direction
            Get
                Return Head
            End Get
            Set(ByVal Value As Direction)
                Head = Value
            End Set
        End Property

        Public Sub WriteCursor(ByVal BW As IO.BinaryWriter)
            BW.Write(MyBase.mTL.TX)
            BW.Write(MyBase.mTL.TY)
            BW.Write(MyBase.Data)
            BW.Write(S1)
            BW.Write(S2)
        End Sub
    End Class
    Public Class MapGrid
        Private Wide As Integer
        Private High As Integer
        Friend TileSize As System.Drawing.Size
        Friend FinalTileSize As System.Drawing.Size
        Private WidthE As Integer
        Private HeightE As Integer
        Public Indicator As MapCursor
        Friend Curspen As System.Drawing.Pen

        Protected mTiles(,) As BackgroundTile

        Public ReadOnly Property Width() As Integer
            Get
                Return Wide
            End Get
        End Property
        Public ReadOnly Property Height() As Integer
            Get
                Return High
            End Get
        End Property
        Public ReadOnly Property FinalColumn() As Integer
            Get
                Return WidthE
            End Get
        End Property
        Public ReadOnly Property FinalRow() As Integer
            Get
                Return HeightE
            End Get
        End Property
        Public Property TileData(ByVal Col As Integer, ByVal Row As Integer) As Short
            Get
                Return mTiles(Col, Row).Data
            End Get
            Set(ByVal Value As Short)
                mTiles(Col, Row).Data = Value
            End Set
        End Property

        Public Sub New(ByVal Width As Integer, ByVal Height As Integer, ByVal TileWidth As Integer, ByVal TileHeight As Integer, ByVal FinalTileWidth As Integer, ByVal FinalTileHeight As Integer)
            Me.Wide = Width
            Me.High = Height
            Me.TileSize = New System.Drawing.Size(TileWidth, TileHeight)
            Me.FinalTileSize = New System.Drawing.Size(FinalTileWidth, FinalTileHeight)
            Me.WidthE = Width - 1
            Me.HeightE = Height - 1
            ReDim mTiles(WidthE, HeightE)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    mTiles(I, J) = New BackgroundTile(I, J)
                Next
            Next
            Me.Indicator = New MapCursor(0, 0, New Rectangle(0, 0, Width, Height))
            Me.Curspen = New System.Drawing.Pen(System.Drawing.Color.Black, 3.0F)
        End Sub
        Public Sub New(ByVal Width As Integer, ByVal Height As Integer, ByVal TileWidth As Integer, ByVal TileHeight As Integer)
            Me.Wide = Width
            Me.High = Height
            Me.TileSize = New System.Drawing.Size(TileWidth, TileHeight)
            Me.FinalTileSize = TileSize
            Me.WidthE = Width - 1
            Me.HeightE = Height - 1
            ReDim mTiles(WidthE, HeightE)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    mTiles(I, J) = New BackgroundTile(I, J)
                Next
            Next
            Me.Indicator = New MapCursor(0, 0, New Rectangle(0, 0, Width, Height))
            Me.Curspen = New System.Drawing.Pen(System.Drawing.Color.Black, 3.0F)
        End Sub
        Public Sub New(ByVal Width As Integer, ByVal Height As Integer, ByVal TileSize As System.Drawing.Size, ByVal FinalTileSize As System.Drawing.Size)
            Me.Wide = Width
            Me.High = Height
            Me.TileSize = TileSize
            Me.FinalTileSize = FinalTileSize
            Me.WidthE = Width - 1
            Me.HeightE = Height - 1
            ReDim mTiles(WidthE, HeightE)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    mTiles(I, J) = New BackgroundTile(I, J)
                Next
            Next
            Me.Indicator = New MapCursor(0, 0, New Rectangle(0, 0, Width, Height))
            Me.Curspen = New System.Drawing.Pen(System.Drawing.Color.Black, 3.0F)
        End Sub
        Public Sub New(ByVal Width As Integer, ByVal Height As Integer, ByVal TileSize As System.Drawing.Size)
            Me.Wide = Width
            Me.High = Height
            Me.TileSize = TileSize
            Me.FinalTileSize = TileSize
            Me.WidthE = Width - 1
            Me.HeightE = Height - 1
            ReDim mTiles(WidthE, HeightE)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    mTiles(I, J) = New BackgroundTile(I, J)
                Next
            Next
            Me.Indicator = New MapCursor(0, 0, New Rectangle(0, 0, Width, Height))
            Me.Curspen = New System.Drawing.Pen(System.Drawing.Color.Black, 3.0F)
        End Sub

        Public Overloads Sub FillMap(ByVal Filename As String)
            Dim FS As System.IO.FileStream = New System.IO.FileStream(Filename, IO.FileMode.Open)
            Dim BR As System.IO.BinaryReader = New System.IO.BinaryReader(FS)
            Dim StrmCatch As Short
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    StrmCatch = BR.ReadInt16()
                    mTiles(I, J).Data = StrmCatch
                Next
            Next
            BR.Close()
            FS.Close()
        End Sub
        Public Overloads Sub FillMap(ByVal BR As IO.BinaryReader)
            Dim StrmCatch As Short
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    StrmCatch = BR.ReadInt16()
                    mTiles(I, J).Data = StrmCatch
                Next
            Next
        End Sub
        Public Overloads Sub FillMap(ByVal Shorts() As Short)
            Dim Ctr As Integer = 0
            If Shorts.GetUpperBound(0) >= (Wide * High - 1) Then
                For J As Integer = 0 To HeightE
                    For I As Integer = 0 To WidthE
                        mTiles(I, J).Data = Shorts(Ctr)
                        Ctr += 1
                    Next
                Next
            Else
                Throw New System.ArgumentOutOfRangeException("Shorts", Shorts.GetUpperBound(0), "Not enough elements in array to fill map.")
            End If
        End Sub
        Public Overloads Sub FillMap(ByVal Data As Short)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    mTiles(I, J).Data = Data
                Next
            Next
        End Sub

        'Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap)
        '    Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
        '    Dim TileData As Short
        '    Dim R As Integer, C As Integer

        '    For J As Integer = 0 To HeightE
        '        For I As Integer = 0 To WidthE
        '            TileData = Me.mTiles(I, J).Data
        '            R = TileSize.Height * (TileData \ NoOfTilesPerRow)
        '            C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
        '            GFX.DrawImage(Tileset, TileSize.Width * I, TileSize.Height * J, New System.Drawing.Rectangle(C, R, TileSize.Width, TileSize.Height), GraphicsUnit.Pixel)
        '        Next
        '    Next
        'End Sub
        Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim TileData As Short
            Dim R, C As Integer

            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData \ NoOfTilesPerRow)
                    C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
                    GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal RedrawZone As System.Drawing.Rectangle)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            LeftBound = RedrawZone.Left \ FinalTileSize.Width
            TopBound = RedrawZone.Top \ FinalTileSize.Height
            RightBound = Convert.ToInt32(System.Math.Ceiling(RedrawZone.Right / FinalTileSize.Width))
            BottomBound = Convert.ToInt32(System.Math.Ceiling(RedrawZone.Right / FinalTileSize.Height))

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData \ NoOfTilesPerRow)
                    C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
                    GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal CursorMoved As Direction)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            Select Case CursorMoved
                Case Direction.Left_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 2, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Direction.Up_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 2, HeightE)
                Case Direction.Right_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 2, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Direction.Down_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 2, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Else
            End Select

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData \ NoOfTilesPerRow)
                    C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
                    GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal ParamArray MapObjects() As MapCharacter)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim TileData As Short
            Dim R, C As Integer

            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData \ NoOfTilesPerRow)
                    C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
                    GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next

            For R = MapObjects.GetLowerBound(0) To MapObjects.GetUpperBound(0)
                With MapObjects(R)
                    GFX.DrawImage(.MyPic, New System.Drawing.Rectangle(.XLocation * FinalTileSize.Width, .YLocation * FinalTileSize.Height, FinalTileSize.Width, FinalTileSize.Height), .FramesToSkip(False), 0, .MyPicFrWidth, .MyPic.Height, GraphicsUnit.Pixel)
                End With
            Next
        End Sub
        Public Overloads Sub DrawMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal RedrawAtTile As TileLocation)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            LeftBound = Math.Max(RedrawAtTile.TX - 1, 0)
            TopBound = Math.Max(RedrawAtTile.TY - 1, 0)
            RightBound = Math.Min(RedrawAtTile.TX + 1, Me.FinalColumn)
            BottomBound = Math.Min(RedrawAtTile.TY + 1, Me.FinalRow)

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData \ NoOfTilesPerRow)
                    C = TileSize.Width * (TileData Mod NoOfTilesPerRow)
                    GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Function GetCloseRectangle(ByVal TL As TileLocation) As Rectangle
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            LeftBound = Math.Max(TL.TX - 2, 0) * FinalTileSize.Width
            TopBound = Math.Max(TL.TY - 2, 0) * FinalTileSize.Height
            RightBound = Math.Min(TL.TX + 2, Me.FinalColumn) * FinalTileSize.Width
            BottomBound = Math.Min(TL.TY + 2, Me.FinalRow) * FinalTileSize.Height
            Return Rectangle.FromLTRB(LeftBound, TopBound, RightBound, BottomBound)
        End Function
        Public Function GetCloseRectangle(ByVal TL As TileLocation, ByVal Moved As Direction) As Rectangle
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            If Moved = Direction.Down_Direction Then
                LeftBound = Math.Max(TL.TX, 0) * FinalTileSize.Width
                TopBound = Math.Max(TL.TY - 2, 0) * FinalTileSize.Height
                RightBound = Math.Min(TL.TX + 1, Me.FinalColumn) * FinalTileSize.Width
                BottomBound = Math.Min(TL.TY + 1, Me.FinalRow) * FinalTileSize.Height
            ElseIf Moved = Direction.Left_Direction Then
                LeftBound = Math.Max(TL.TX, 0) * FinalTileSize.Width
                TopBound = Math.Max(TL.TY - 1, 0) * FinalTileSize.Height
                RightBound = Math.Min(TL.TX + 1, Me.FinalColumn) * FinalTileSize.Width
                BottomBound = Math.Min(TL.TY + 1, Me.FinalRow) * FinalTileSize.Height
            ElseIf Moved = Direction.Right_Direction Then
                LeftBound = Math.Max(TL.TX - 1, 0) * FinalTileSize.Width
                TopBound = Math.Max(TL.TY - 1, 0) * FinalTileSize.Height
                RightBound = Math.Min(TL.TX + 1, Me.FinalColumn) * FinalTileSize.Width
                BottomBound = Math.Min(TL.TY + 1, Me.FinalRow) * FinalTileSize.Height
            ElseIf Moved = Direction.Up_Direction Then
                LeftBound = Math.Max(TL.TX, 0) * FinalTileSize.Width
                TopBound = Math.Max(TL.TY - 1, 0) * FinalTileSize.Height
                RightBound = Math.Min(TL.TX + 1, Me.FinalColumn) * FinalTileSize.Width
                BottomBound = Math.Min(TL.TY + 2, Me.FinalRow) * FinalTileSize.Height
            End If

            Return Rectangle.FromLTRB(LeftBound, TopBound, RightBound, BottomBound)
        End Function


        Public Overloads Sub DrawCursor(ByRef GFX As System.Drawing.Graphics, ByVal Color As System.Drawing.Color)
            Me.Curspen.Color = Color
            GFX.DrawRectangle(Me.Curspen, New Rectangle(New System.Drawing.Point(FinalTileSize.Width * Me.Indicator.XLocation, FinalTileSize.Height * Me.Indicator.YLocation), FinalTileSize))
        End Sub
        Public Overloads Sub DrawCursor(ByRef GFX As System.Drawing.Graphics)
            'Me.Curspen.Color = Color
            GFX.DrawRectangle(Me.Curspen, New Rectangle(New System.Drawing.Point(FinalTileSize.Width * Me.Indicator.XLocation, FinalTileSize.Height * Me.Indicator.YLocation), FinalTileSize))
        End Sub


        'The 16 versions are specially designed functions for tilesize and finaltilesize of 16x16
        'number of tiles per row = 16
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap)
            Dim TileData As Short
            Dim R, C As Integer

            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(I << 4, J << 4), FinalTileSize), C, R, 16, 16, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal RedrawZone As System.Drawing.Rectangle)
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            LeftBound = RedrawZone.Left >> 4
            TopBound = RedrawZone.Top >> 4
            If (RedrawZone.Right And &HF) = 0 Then
                RightBound = RedrawZone.Right >> 4
            Else
                RightBound = (RedrawZone.Right + 1) >> 4
            End If
            If (RedrawZone.Bottom And &HF) = 0 Then
                BottomBound = RedrawZone.Bottom >> 4
            Else
                BottomBound = (RedrawZone.Bottom + 1) >> 4
            End If

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal CursorMoved As Direction)

            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            Select Case CursorMoved
                Case Direction.Left_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 2, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Direction.Up_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 2, HeightE)
                Case Direction.Right_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 2, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 1, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Direction.Down_Direction
                    LeftBound = System.Math.Max(Indicator.XLocation - 1, 0)
                    TopBound = System.Math.Max(Indicator.YLocation - 2, 0)
                    RightBound = System.Math.Min(Indicator.XLocation + 1, WidthE)
                    BottomBound = System.Math.Min(Indicator.YLocation + 1, HeightE)
                Case Else
            End Select

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal ParamArray MapObjects() As MapCharacter)
            Dim TileData As Short
            Dim R, C As Integer

            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next

            For R = MapObjects.GetLowerBound(0) To MapObjects.GetUpperBound(0)
                With MapObjects(R)
                    GFX.DrawImage(.MyPic, New System.Drawing.Rectangle(.XLocation << 4, .YLocation << 4, 16, 16), .FramesToSkip(False), 0, .MyPicFrWidth, .MyPic.Height, GraphicsUnit.Pixel)
                End With
            Next
        End Sub
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal RedrawAtTile As TileLocation)
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            LeftBound = Math.Max(RedrawAtTile.TX - 1, 0)
            TopBound = Math.Max(RedrawAtTile.TY - 1, 0)
            RightBound = Math.Min(RedrawAtTile.TX + 1, Me.FinalColumn)
            BottomBound = Math.Min(RedrawAtTile.TY + 1, Me.FinalRow)

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Overloads Sub DrawMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal RedrawAtTile As TileLocation, ByVal CursorMoved As Direction)
            Dim TileData As Short
            Dim R, C As Integer
            Dim LeftBound, RightBound, TopBound, BottomBound As Integer
            If CursorMoved = Direction.Down_Direction Then
                LeftBound = Math.Max(RedrawAtTile.TX, 0)
                TopBound = Math.Max(RedrawAtTile.TY - 2, 0)
                RightBound = Math.Min(RedrawAtTile.TX, Me.FinalColumn)
                BottomBound = Math.Min(RedrawAtTile.TY, Me.FinalRow)
            ElseIf CursorMoved = Direction.Left_Direction Then
                LeftBound = Math.Max(RedrawAtTile.TX, 0)
                TopBound = Math.Max(RedrawAtTile.TY - 1, 0)
                RightBound = Math.Min(RedrawAtTile.TX + 1, Me.FinalColumn)
                BottomBound = Math.Min(RedrawAtTile.TY, Me.FinalRow)
            ElseIf CursorMoved = Direction.Right_Direction Then
                LeftBound = Math.Max(RedrawAtTile.TX - 1, 0)
                TopBound = Math.Max(RedrawAtTile.TY - 1, 0)
                RightBound = Math.Min(RedrawAtTile.TX, Me.FinalColumn)
                BottomBound = Math.Min(RedrawAtTile.TY, Me.FinalRow)
            ElseIf CursorMoved = Direction.Up_Direction Then
                LeftBound = Math.Max(RedrawAtTile.TX, 0)
                TopBound = Math.Max(RedrawAtTile.TY - 1, 0)
                RightBound = Math.Min(RedrawAtTile.TX, Me.FinalColumn)
                BottomBound = Math.Min(RedrawAtTile.TY + 1, Me.FinalRow)
            End If

            For J As Integer = TopBound To BottomBound
                For I As Integer = LeftBound To RightBound
                    TileData = Me.mTiles(I, J).Data
                    R = TileSize.Height * (TileData >> 4)
                    C = TileSize.Width * (TileData And &HF)
                    GFX.DrawImage(Tileset, I << 4, J << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                    'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * I, FinalTileSize.Height * J), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)
                Next
            Next
        End Sub
        Public Sub PlaceCursorData()
            Me.mTiles(Me.Indicator.XLocation, Me.Indicator.YLocation).Data = Me.Indicator.Data
        End Sub

        Public Overloads Sub DrawOnMap(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation)
            Dim DrawX As Integer = TileLoc.TX * Me.TileSize.Width
            Dim DrawY As Integer = TileLoc.TY * Me.TileSize.Height
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation, ByVal InSet As System.Drawing.Size)
            Dim DrawX As Integer = TileLoc.TX * Me.TileSize.Width - InSet.Width
            Dim DrawY As Integer = TileLoc.TY * Me.TileSize.Height - InSet.Height
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation, ByVal VerticalInset As Integer)
            Dim DrawX As Integer = TileLoc.TX * Me.TileSize.Width
            Dim DrawY As Integer = TileLoc.TY * Me.TileSize.Height - VerticalInset
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal Emblem As MapCursor)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim DrawX As Integer = Emblem.XLocation * Me.TileSize.Width
            Dim DrawY As Integer = Emblem.YLocation * Me.TileSize.Height
            Dim R, C As Integer

            R = TileSize.Height * (Emblem.Data \ NoOfTilesPerRow)
            C = TileSize.Width * (Emblem.Data Mod NoOfTilesPerRow)

            GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * Emblem.XLocation, FinalTileSize.Height * Emblem.YLocation), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)

        End Sub
        Public Overloads Sub DrawOnMap(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal Emblems As MapCursorSet)
            Dim NoOfTilesPerRow As Integer = Tileset.Width \ TileSize.Width
            Dim R, C As Integer
            Dim Emblem As MapCursor

            For LV As Integer = 0 To Emblems.Count - 1
                Emblem = Emblems(LV, False)
                Dim DrawX As Integer = Emblem.XLocation * Me.TileSize.Width
                Dim DrawY As Integer = Emblem.YLocation * Me.TileSize.Height

                R = TileSize.Height * (Emblem.Data \ NoOfTilesPerRow)
                C = TileSize.Width * (Emblem.Data Mod NoOfTilesPerRow)

                GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * Emblem.XLocation, FinalTileSize.Height * Emblem.YLocation), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)

            Next


        End Sub
        Public Overloads Sub DrawOnMap16(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation)
            Dim DrawX As Integer = TileLoc.TX << 4
            Dim DrawY As Integer = TileLoc.TY << 4
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap16(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation, ByVal InSet As System.Drawing.Size)
            Dim DrawX As Integer = (TileLoc.TX << 4) - InSet.Width
            Dim DrawY As Integer = (TileLoc.TY << 4) - InSet.Height
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap16(ByRef GFX As System.Drawing.Graphics, ByRef Bmp As System.Drawing.Bitmap, ByVal TileLoc As TileLocation, ByVal VerticalInset As Integer)
            Dim DrawX As Integer = (TileLoc.TX << 4)
            Dim DrawY As Integer = (TileLoc.TY << 4) - VerticalInset
            GFX.DrawImage(Bmp, DrawX, DrawY)
        End Sub
        Public Overloads Sub DrawOnMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal Emblem As MapCursor)
            Dim DrawX As Integer = Emblem.XLocation << 4
            Dim DrawY As Integer = Emblem.YLocation << 4
            Dim R, C As Integer

            R = TileSize.Height * (Emblem.Data >> 4)
            C = TileSize.Width * (Emblem.Data And &HF)

            GFX.DrawImage(Tileset, Emblem.XLocation << 4, Emblem.YLocation << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)

            'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * Emblem.XLocation, FinalTileSize.Height * Emblem.YLocation), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)

        End Sub
        Public Overloads Sub DrawOnMap16(ByRef GFX As System.Drawing.Graphics, ByRef Tileset As System.Drawing.Bitmap, ByVal Emblems As MapCursorSet)
            Dim R, C As Integer
            Dim Emblem As MapCursor

            For LV As Integer = 0 To Emblems.Count - 1
                Emblem = Emblems(LV, False)
                Dim DrawX As Integer = Emblem.XLocation << 4
                Dim DrawY As Integer = Emblem.YLocation << 4

                R = TileSize.Height * (Emblem.Data >> 4)
                C = TileSize.Width * (Emblem.Data And &HF)

                GFX.DrawImage(Tileset, Emblem.XLocation << 4, Emblem.YLocation << 4, New Rectangle(C, R, 16, 16), GraphicsUnit.Pixel)
                'GFX.DrawImage(Tileset, New System.Drawing.Rectangle(New System.Drawing.Point(FinalTileSize.Width * Emblem.XLocation, FinalTileSize.Height * Emblem.YLocation), FinalTileSize), C, R, TileSize.Width, TileSize.Height, GraphicsUnit.Pixel)

            Next


        End Sub

        Public Sub ShiftToCleanMap(ByVal F As Direction)
            Dim LV, IV As Integer
            If F = Direction.Left_Direction Then
                For LV = 0 To High
                    mTiles(Wide, LV).Data = mTiles(0, LV).Data
                    For IV = 0 To Wide - 1
                        mTiles(IV, LV).Data = Me.Indicator.Data
                    Next
                Next
            ElseIf F = Direction.Up_Direction Then
                For LV = 0 To Wide
                    mTiles(LV, High).Data = mTiles(LV, 0).Data
                    For IV = 0 To High - 1
                        mTiles(LV, IV).Data = Me.Indicator.Data
                    Next
                Next
            ElseIf F = Direction.Right_Direction Then
                For LV = 0 To High
                    mTiles(0, LV).Data = mTiles(Wide, LV).Data
                    For IV = 1 To Wide
                        mTiles(IV, LV).Data = Me.Indicator.Data
                    Next
                Next
            ElseIf F = Direction.Down_Direction Then
                For LV = 0 To Wide
                    mTiles(LV, 0).Data = mTiles(LV, High).Data
                    For IV = 1 To High
                        mTiles(LV, IV).Data = Me.Indicator.Data
                    Next
                Next

            End If
        End Sub


        Public Sub MapToFile(ByVal Filename As String)
            Dim FS As System.IO.FileStream = New System.IO.FileStream(Filename, IO.FileMode.Create)
            Dim BW As System.IO.BinaryWriter = New System.IO.BinaryWriter(FS)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    BW.Write(Me.mTiles(I, J).Data)
                Next
            Next
            BW.Close()
            FS.Close()
        End Sub
        Public Sub MapToFile(ByVal BW As IO.BinaryWriter)
            For J As Integer = 0 To HeightE
                For I As Integer = 0 To WidthE
                    BW.Write(Me.mTiles(I, J).Data)
                Next
            Next
        End Sub

        Public Function TileFromPoint(ByVal X1 As Integer, ByVal Y1 As Integer) As BackgroundTile
            Dim Tx As Integer = X1 \ FinalTileSize.Width
            Dim Ty As Integer = Y1 \ FinalTileSize.Height

            If Tx >= 0 AndAlso Tx <= Me.FinalColumn AndAlso Ty >= 0 AndAlso Ty <= Me.FinalRow Then
                Return Me.mTiles(Tx, Ty)
            Else
                Return Nothing
            End If
        End Function

        Public Function GetTilesOfLine(ByVal X1 As Single, ByVal Y1 As Single, ByVal X2 As Single, ByVal Y2 As Single, ByVal RemoveEndPoints As Boolean) As BackgroundTile()
            Dim AL As ArrayList = New ArrayList(40)
            Dim T As BackgroundTile = Me.TileFromPoint(Convert.ToInt32(X1), Convert.ToInt32(Y1))
            Dim F As BackgroundTile = Me.TileFromPoint(Convert.ToInt32(X2), Convert.ToInt32(Y2))
            Dim Slope As Single = Convert.ToSingle((Y2 - Y1) / (X2 - X1))
            Dim SX As Single = X1
            Dim SY As Single = Y1

            Dim DistX, DistY As Single
            Dim GoingRight As Boolean = X2 >= X1
            If GoingRight Then
                Do Until T.Tilelocation.Equals(F.Tilelocation)
                    AL.Add(T)
                    DistX = FinalTileSize.Width - FloatMod(SX, FinalTileSize.Width)
                    If Slope < 0 Then
                        DistY = -FloatMod(SY, FinalTileSize.Height)
                        If DistY = 0.0F Then
                            DistY = -FinalTileSize.Height
                        End If
                    Else
                        DistY = FinalTileSize.Height - FloatMod(SY, FinalTileSize.Height)
                    End If

                    If Math.Abs(Slope * DistX) > Math.Abs(DistY) Then
                        SY += DistY
                        SX += DistY / Slope
                    Else
                        SX += DistX
                        SY += DistX * Slope
                    End If
                    If Slope < 0 AndAlso FloatMod(SY, FinalTileSize.Height) = 0 Then
                        T = Me.TileFromPoint(Convert.ToInt32(SX), Convert.ToInt32(SY - FinalTileSize.Height / 2))
                    Else
                        T = Me.TileFromPoint(Convert.ToInt32(SX), Convert.ToInt32(SY))
                    End If
                Loop
            Else
                Do Until T.Tilelocation.Equals(F.Tilelocation)
                    AL.Add(T)
                    DistX = FloatMod(SX, FinalTileSize.Width)
                    If DistX = 0.0F Then
                        DistX = FinalTileSize.Width
                    End If
                    If Slope > 0 Then
                        DistY = -FloatMod(SY, FinalTileSize.Height)
                        If DistY = 0.0F Then
                            DistY = -FinalTileSize.Height
                        End If
                    Else
                        DistY = FinalTileSize.Height - FloatMod(SY, FinalTileSize.Height)
                    End If

                    If Math.Abs(Slope * DistX) > Math.Abs(DistY) Then
                        SY += DistY
                        SX += DistY / Slope
                    Else
                        SX -= DistX
                        SY -= DistX * Slope
                    End If
                    If Slope > 0 AndAlso FloatMod(SY, FinalTileSize.Height) = 0 Then
                        If FloatMod(SX, FinalTileSize.Width) = 0 Then
                            T = Me.TileFromPoint(Convert.ToInt32(SX - FinalTileSize.Width / 2), Convert.ToInt32(SY - FinalTileSize.Height / 2))
                        Else
                            T = Me.TileFromPoint(Convert.ToInt32(SX), Convert.ToInt32(SY - FinalTileSize.Height / 2))

                        End If
                    Else
                        If FloatMod(SX, FinalTileSize.Width) = 0 Then
                            T = Me.TileFromPoint(Convert.ToInt32(SX - FinalTileSize.Width / 2), Convert.ToInt32(SY))
                        Else
                            T = Me.TileFromPoint(Convert.ToInt32(SX), Convert.ToInt32(SY))

                        End If
                    End If
                Loop

            End If

            If RemoveEndPoints Then
                If AL.Count = 1 Then
                    AL.RemoveAt(0)
                ElseIf AL.Count > 1 Then
                    AL.RemoveAt(AL.Count - 1)
                    AL.RemoveAt(0)
                End If
            End If

            Dim A As Array = AL.ToArray(T.GetType())
            Return DirectCast(A, BackgroundTile())

        End Function

        Public Function GetTLMidPoint(ByVal TL As TileLocation) As Point
            Return New Point(TL.TX * FinalTileSize.Width + (FinalTileSize.Width >> 1), TL.TY * FinalTileSize.Height + (FinalTileSize.Height >> 1))
        End Function

        Public Property TileData(ByVal MC As MapCursor) As Short
            Set(ByVal Value As Short)
                Me.mTiles(MC.XLocation(), MC.YLocation()).Data = Value
            End Set
            Get
                Return Me.mTiles(MC.XLocation(), MC.YLocation()).Data
            End Get
        End Property

        Public Function IndexOf(ByVal Dat As Short) As BackgroundTile

            Dim X, Y As Integer, BT As BackgroundTile
            Do Until (Y >= Height) OrElse Not (BT Is Nothing)

                If mTiles(X, Y).Data = Dat Then
                    BT = mTiles(X, Y)
                End If

                X += 1
                If X >= Width Then
                    X = 0
                    Y += 1
                End If
            Loop
            Return BT
        End Function
        Public Function IndexOf(ByVal Dat As Short, ByVal Skip As TileLocation) As BackgroundTile

            Dim X, Y As Integer, BT As BackgroundTile
            Do Until (X >= Width AndAlso Y >= Height) OrElse Not (BT Is Nothing)

                If mTiles(X, Y).Data = Dat AndAlso Not (Skip.TX = X AndAlso Skip.TY = Y) Then
                    BT = mTiles(X, Y)
                End If

                X += 1
                If X >= Width Then
                    X = 0
                    Y += 1
                End If
            Loop
            Return BT
        End Function
        Public Function IndexOf(ByVal StartAt As TileLocation, ByVal Dat As Short) As BackgroundTile
            Dim X, Y As Integer, BT As BackgroundTile
            X = StartAt.TX
            Y = StartAt.TY
            Do Until (X >= Width AndAlso Y >= Height) OrElse Not (BT Is Nothing)

                If mTiles(X, Y).Data = Dat Then
                    BT = mTiles(X, Y)
                End If

                X += 1
                If X >= Width Then
                    X = 0
                    Y += 1
                End If
            Loop
            Return BT
        End Function

        Private Function FloatMod(ByVal Dividend As Single, ByVal Divisor As Single) As Single
            Return Convert.ToSingle((Dividend / Divisor - Math.Floor(Dividend / Divisor)) * Divisor)
        End Function
    End Class
    Public Enum MapCharPicFormats
        SinglePic = 0
        FourDirectionPics
        FourDirectionWithStep
    End Enum
    Public Class MapCharacter
        Inherits MapCursor
        Friend MyPic As System.Drawing.Bitmap
        Friend MyPicFormat As MapCharPicFormats
        Friend MyPicFrWidth As Integer
        Public Direc As Direction

        Public Sub New(ByVal StartX As Integer, ByVal StartY As Integer, ByRef Boundary As System.Drawing.Rectangle, ByRef Pic As System.Drawing.Bitmap, ByVal PicFormat As MapCharPicFormats, ByVal PicFrameWidth As Integer)
            MyBase.New(StartX, StartY, Boundary)
            MyPic = Pic
            MyPicFormat = PicFormat
            '= 1 To 1
        End Sub

        Friend Function FramesToSkip(ByVal DoStep As Boolean) As Integer
            Select Case MyPicFormat
                Case MapCharPicFormats.SinglePic
                    Return 0
                Case MapCharPicFormats.FourDirectionPics
                    Return Direc * MyPicFrWidth
                Case MapCharPicFormats.FourDirectionWithStep
                    If DoStep Then
                        Return (2 * Direc + 1) * MyPicFrWidth
                    Else
                        Return 2 * Direc * MyPicFrWidth
                    End If
                Case Else
            End Select
        End Function
    End Class
    Public Class DirectionEventArgs
        Inherits System.EventArgs
        Public D As Direction
        Public Sub New(ByVal Direction As Direction)
            D = Direction
        End Sub
    End Class
    Public Class CursorCollisionEventArgs
        Inherits System.EventArgs
        Public ColliderIndex As Integer
        Public CollideeIndex As Integer
        Public Sub New(ByVal ColliderIndex As Integer, ByVal CollideeIndex As Integer)
            Me.CollideeIndex = CollideeIndex
            Me.ColliderIndex = ColliderIndex
        End Sub
    End Class
    Public Delegate Sub DirectionEventHandler(ByVal sender As Object, ByVal e As DirectionEventArgs)
    Public Delegate Sub CursorCollisionEventHandler(ByVal sender As Object, ByVal e As CursorCollisionEventArgs)
    Public Class MapCursorSet
        Private Cursors As ArrayList

        Public Event Collision As CursorCollisionEventHandler
        Public Sub New(ByVal Capacity As Integer)
            Cursors = New ArrayList(Capacity)
        End Sub

        Public Sub Add(ByVal Item As MapCursor)
            Cursors.Add(Item)
        End Sub

        Default Public Property Item(ByVal Index As Integer, ByVal CollisionCheck As Boolean) As MapCursor
            Get
                Return DirectCast(Cursors(Index), MapCursor)
            End Get
            Set(ByVal Value As MapCursor)
                Dim Curse As MapCursor, LV As Integer, dSize As Integer
                Cursors(Index) = Value
                If CollisionCheck Then
                    LV = 0
                    Do Until LV = Cursors.Count
                        dSize = 0
                        If Index <> LV Then
                            Curse = DirectCast(Cursors(LV), MapCursor)
                            If Curse.Tilelocation.Equals(Value.Tilelocation) AndAlso Not (Curse Is Value) Then
                                dSize = Cursors.Count
                                RaiseEvent Collision(Me, New CursorCollisionEventArgs(Index, LV))
                                dSize -= Cursors.Count
                            End If
                        End If
                        LV += (1 - dSize)
                    Loop
                End If
            End Set
        End Property

        Public Function IndexOf(ByVal Curse As MapCursor) As Integer
            Return Cursors.IndexOf(Curse)
        End Function

        Public Function IndexOfData(ByVal S2 As Short) As Integer
            Dim Curse As MapCursor, LV As Integer
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors.Item(LV), MapCursor)
                If Curse.Data = S2 Then
                    IndexOfData = LV
                    LV = Cursors.Count
                End If
            Next
        End Function
        Public Function CursorAtData(ByVal S2 As Short) As MapCursor
            Dim Curse As MapCursor, LV As Integer
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors.Item(LV), MapCursor)
                If Curse.Data = S2 Then
                    CursorAtData = Curse
                    LV = Cursors.Count
                End If
            Next
        End Function
        Public Function CursorsAtData(ByVal S2 As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor, LV As Integer
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors.Item(LV), MapCursor)
                If Curse.Data = S2 Then
                    AL.Add(Curse)
                End If
            Next
            If Curse Is Nothing Then
                Return New MapCursor() {}
            Else
                Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
            End If

        End Function

        Public ReadOnly Property Count() As Integer
            Get

                Return Cursors.Count
            End Get
        End Property

        Public Function Contains(ByVal Curse As MapCursor) As Boolean
            Return Cursors.Contains(Curse)
        End Function

        Public Function CursorAtLocation(ByVal TL As TileLocation) As Integer
            Dim LV As Integer, Curse As MapCursor
            CursorAtLocation = -1
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) Then
                    CursorAtLocation = LV
                    LV = Cursors.Count
                End If
            Next
        End Function
        Public Function CursorAtLocation(ByVal TL As TileLocation, ByVal Skip As Integer) As Integer
            Dim LV As Integer, Curse As MapCursor
            CursorAtLocation = -1
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso LV <> Skip Then
                    CursorAtLocation = LV
                    LV = Cursors.Count
                End If
            Next
        End Function
        Public Function CursorAtLocation(ByVal TL As TileLocation, ByVal SkipDataArr() As Short) As Integer
            Dim LV As Integer, Curse As MapCursor
            CursorAtLocation = -1
            For LV = 0 To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso Array.IndexOf(SkipDataArr, Curse.Data) = -1 Then
                    CursorAtLocation = LV
                    LV = Cursors.Count
                End If
            Next
        End Function

        Public Function CursorAtLocation(ByVal Start As Integer, ByVal TL As TileLocation) As Integer
            Dim LV As Integer, Curse As MapCursor
            CursorAtLocation = -1
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) Then
                    CursorAtLocation = LV
                    LV = Cursors.Count
                End If
            Next
        End Function
        Public Function CursorAtLocation(ByVal Start As Integer, ByVal TL As TileLocation, ByVal Skipdataarr() As Short) As Integer
            Dim LV As Integer, Curse As MapCursor
            CursorAtLocation = -1
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso Array.IndexOf(Skipdataarr, Curse.Data) = -1 Then
                    CursorAtLocation = LV
                    LV = Cursors.Count
                End If
            Next
        End Function

        Public Sub Remove(ByVal MC As MapCursor)
            Cursors.Remove(MC)
        End Sub

        Public Sub RemoveAt(ByVal Index As Integer)
            Cursors.RemoveAt(Index)
        End Sub

        Public Function CursorsAtPosition(ByVal BT As BackgroundTile) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                If Curse.Tilelocation.Equals(BT.Tilelocation()) Then
                    AL.Add(Curse)
                End If
            Next
            If Curse Is Nothing Then
                Return New MapCursor() {}
            Else
                Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
            End If
        End Function
        Public Function CursorsAtPosition(ByVal TL As TileLocation) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                If Curse.Tilelocation.Equals(TL) Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal TL As TileLocation, ByVal SkipData As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                If Curse.Tilelocation.Equals(TL) AndAlso Curse.Data <> SkipData Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal TL As TileLocation, ByVal SkipData1 As Short, ByVal SkipData2 As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                If Curse.Tilelocation.Equals(TL) AndAlso Curse.Data <> SkipData1 AndAlso Curse.Data <> SkipData2 Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal TL As TileLocation, ByVal SkipArr() As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                If Curse.Tilelocation.Equals(TL) AndAlso Array.IndexOf(SkipArr, Curse.Data) = -1 Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal Start As Integer, ByVal TL As TileLocation) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor, LV As Integer
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal Start As Integer, ByVal TL As TileLocation, ByVal SkipData As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor, LV As Integer
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso Curse.Data <> SkipData Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal Start As Integer, ByVal TL As TileLocation, ByVal SkipData1 As Short, ByVal SkipData2 As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor, LV As Integer
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso Curse.Data <> SkipData1 AndAlso Curse.Data <> SkipData2 Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function
        Public Function CursorsAtPosition(ByVal Start As Integer, ByVal TL As TileLocation, ByVal SkipArr() As Short) As MapCursor()
            Dim AL As ArrayList = New ArrayList(4)
            Dim Curse As MapCursor, LV As Integer
            For LV = Start To Cursors.Count - 1
                Curse = DirectCast(Cursors(LV), MapCursor)
                If Curse.Tilelocation.Equals(TL) AndAlso Array.IndexOf(SkipArr, Curse.Data) = -1 Then
                    AL.Add(Curse)
                End If
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())
        End Function


        Public Function CursorsFromTilelist(ByVal BTS() As BackgroundTile) As MapCursor()
            Dim AL As ArrayList = New ArrayList(15)
            Dim Curse As MapCursor
            For Each Curse In Cursors
                For LV As Integer = 0 To BTS.GetUpperBound(0)
                    If Curse.Tilelocation.Equals(BTS(LV).Tilelocation()) Then
                        AL.Add(Curse)
                    End If
                Next
            Next

            Return DirectCast(AL.ToArray(Curse.GetType()), MapCursor())

        End Function

        Public Sub AddRange(ByVal Items() As MapCursor)
            Cursors.AddRange(Items)
        End Sub

        Public Sub UpdateByIndexReassign(ByVal MC As MapCursor)
            Dim Idx As Integer = Cursors.IndexOf(MC)
            Me.Item(Idx, True) = MC
        End Sub

        Public Sub Clear()
            Cursors.Clear()
        End Sub

        Public Sub Sort()
            Cursors.Sort()
        End Sub
    End Class

End Namespace
